#!/usr/bin/env python
import pytest
from hypothesis import given
import hypothesis.strategies as st
from tibs import Tibs, Mutibs
from typing import Iterable


class TestCreation:
    def test_creation_from_bytes(self):
        s = Tibs.from_bytes(b"\xa0\xff")
        assert (len(s), s.to_hex()) == (16, "a0ff")

    @given(st.binary())
    def test_creation_from_bytes_roundtrip(self, data):
        s = Tibs.from_bytes(data)
        assert s.to_bytes() == data

    def test_creation_from_hex(self):
        s = Tibs.from_hex("0xA0ff")
        assert (len(s), s.to_hex()) == (16, "a0ff")


class TestInitialisation:
    def test_empty_init(self):
        a = Tibs()
        assert a == ""

    def test_find(self):
        a = Tibs.from_string("0xabcd")
        r = a.find("0xbc")
        assert r == 4
        r = a.find("0x23462346246", byte_aligned=True)
        assert r is None

    def test_rfind(self):
        a = Tibs.from_string("0b11101010010010")
        b = a.rfind("0b010")
        assert b == 11

    def test_find_all(self):
        a = Tibs("0b0010011")
        b = list(a.find_all('0b1'))
        assert b == [2, 5, 6]
        t = Tibs("0b10")
        tp = list(t.find_all("0b1"))
        assert tp == [0]


class TestCut:
    def test_cut(self):
        s = Tibs().from_joined(["0b000111"] * 10)
        for t in s.chunks(6):
            assert t == "0b000111"


def test_unorderable():
    a = Tibs("0b000111")
    b = Tibs("0b000111")
    with pytest.raises(TypeError):
        _ = a < b
    with pytest.raises(TypeError):
        _ = a > b
    with pytest.raises(TypeError):
        _ = a <= b
    with pytest.raises(TypeError):
        _ = a >= b


class TestPadToken:
    def test_creation(self):
        with pytest.raises(ValueError):
            _ = Tibs.from_string("pad10")
        with pytest.raises(ValueError):
            _ = Tibs.from_string("pad")


def test_adding():
    a = Tibs.from_string("0b0")
    b = Tibs.from_string("0b11")
    c = a + b
    assert c == "0b011"
    assert a == "0b0"
    assert b == "0b11"


class TestContainsBug:
    def test_contains(self):
        a = Tibs.from_string("0b1, 0x0001dead0001")
        assert "0xdead" in a
        assert "0xfeed" not in a

        assert "0b1" in Tibs.from_string("0xf")
        assert "0b0" not in Tibs.from_string("0xf")


class TestUnderscoresInLiterals:
    def test_hex_creation(self):
        a = Tibs.from_hex("ab_cd__ef")
        assert a.to_hex() == "abcdef"
        b = Tibs.from_string("0x0102_0304")
        assert b.to_hex() == "01020304"

    def test_binary_creation(self):
        a = Tibs.from_bin("0000_0001_0010")
        assert a.to_bin() == "000000010010"
        b = Tibs.from_string("0b0011_1100_1111_0000")
        assert b.to_bin() == "0011110011110000"

    def test_octal_creation(self):
        a = Tibs.from_oct("0011_2233_4455_6677")
        assert a.to_oct() == "0011223344556677"
        b = Tibs.from_string("0o123_321_123_321")
        assert b.to_oct() == "123321123321"


def test_from_iterable():
    with pytest.raises(TypeError):
        _ = Tibs.from_bools()
    a = Tibs.from_bools([])
    assert a == Tibs()
    a = Tibs.from_bools([1, 0, 1, 1])
    assert a == "0b1011"
    a = Tibs.from_bools((True,))
    assert a.to_bin() == "1"


def test_mul_by_zero():
    a = Tibs.from_string("0b1010")
    b = a * 0
    assert b == Tibs()
    b = a * 1
    assert b == a
    b = a * 2
    assert b == a + a


def test_from_ones():
    a = Tibs.from_ones(0)
    assert a == Tibs()
    a = Tibs.from_ones(1)
    assert a == Tibs("0b1")
    with pytest.raises(ValueError):
        _ = Tibs.from_ones(-1)


def test_from_zeros():
    a = Tibs.from_zeros(0)
    assert a == Tibs()
    a = Tibs.from_zeros(1)
    assert a == Tibs("0b0")
    with pytest.raises(ValueError):
        _ = Tibs.from_zeros(-1)


def test_bits_slicing():
    a = Tibs('0b1010101010101010')
    b = a[-5:-8:1]
    assert b == Tibs()

    assert a[::2] == '0xff'
    assert a[1::2] == '0x00'


def test_from_random():
    a = Tibs.from_random(0)
    assert a == Tibs()
    a = Tibs.from_random(1)
    assert a == '0b1' or a == '0b0'
    a = Tibs.from_random(10000, seed=b'a_seed')
    b = Tibs.from_random(10000, seed=b'a_seed')
    assert a == b
    b = Tibs.from_random(10000,
                         seed=b'a different seed this time - quite long to test if this makes a difference or not. It shouldnt really, but who knows?')
    assert a != b
    c = Mutibs.from_random(10000, seed=b'a_seed')
    assert a == c


def test_is_things():
    a = Tibs('0b1010101010101010')
    b = Mutibs('0b1')
    assert isinstance(a, Iterable)
    assert isinstance(b, Iterable)


def test_bits_from_bytes_string():
    a = Tibs.from_bytes(b'ABC')
    assert a.to_bytes() == b'ABC'


def test_bool_conversion():
    a = Tibs()
    b = Tibs('0b0')
    c = Tibs('0b1')
    assert not a
    assert b
    assert c


def test_find_all():
    a = Tibs(' 0 B 0 0 01011')
    g = a.find_all('0b1')
    assert next(g) == 3
    assert next(g) == 5
    assert next(g) == 6
    with pytest.raises(StopIteration):
        _ = next(g)


def test_repr():
    a = Tibs()
    assert repr(a) == "Tibs()"
    a = Tibs('')
    assert repr(a) == "Tibs()"
    a = Tibs(" 0b 1")
    assert repr(a) == "Tibs('0b1')"


def test_bits_not_orderable():
    a = Tibs.from_string("0b0")
    b = Tibs.from_string("0b1")
    with pytest.raises(TypeError):
        _ = a < b
    with pytest.raises(TypeError):
        _ = a <= b
    with pytest.raises(TypeError):
        _ = a > b
    with pytest.raises(TypeError):
        _ = a >= b


def test_bools_from_iterable():
    v = [1, 0, 0, 1]
    i = iter(v)
    b = Tibs.from_bools(i)
    assert b == '0b1001'


def test_joined_from_iterable():
    v = [[0], '0b11']
    i = iter(v)
    b = Tibs.from_joined(v)
    assert b == '0b011'



from tibs import Mutibs
from math import isqrt

# Create a billion True bits
limit = 1_000_000_000
is_prime = Mutibs.from_ones(limit)

# Zero and one aren't prime, so set these to False
is_prime.set(False, [0, 1])

# Set all bits that are a multiple of the lowest known prime to False
for i in range(2, isqrt(limit) + 1):
    if is_prime[i]:
        is_prime.set(False, range(i * i, limit, i))

# We can now use it to count the primes less than a billion
primes_count = is_prime.count(1)
assert primes_count == 50_847_534

# Let's also see how many twin primes there are (primes that differ by 2).
twin_primes = len(list(is_prime.as_tibs().find_all('0b101')))
assert twin_primes == 239_101



#!/usr/bin/env python
import pytest
from tibs import Tibs, Mutibs


def test_creation():
    a = Mutibs.from_zeros(5)
    b = Mutibs.from_bools([1, 0, 0])
    c = Mutibs.from_bytes(b'123')
    e = Mutibs.from_string('0b1110')
    for x in [a, b, c, e]:
        assert isinstance(x, Mutibs)


def test_append():
    # Basic append functionality
    a = Mutibs('0x0f')
    a.append('0x0a')
    assert a == '0x0f0a'

    # Verify method chaining
    a = Mutibs('0x01')
    result = a.append('0x02').append('0x03')
    assert a == '0x010203'
    assert result is a  # Should return self

    # Different input types
    a = Mutibs('0b1010')
    a.append(Tibs('0b1111'))  # Tibs object
    assert a == '0b10101111'
    a.append(Tibs.from_bools([True, False, True]))
    assert a == '0b10101111101'

    # Empty append
    a = Mutibs('0x42')
    a.append(Tibs())
    assert a == '0x42'


def test_prepend():
    # Basic prepend functionality
    a = Mutibs('0x0f')
    a.prepend('0x0a')
    assert a == '0x0a0f'

    # Verify method chaining
    a = Mutibs('0x03')
    result = a.prepend('0x02').prepend('0x01')
    assert a == '0x010203'
    assert result is a  # Should return self

    # Different input types
    a = Mutibs('0b1010')
    a.prepend(Tibs('0b1111'))  # Tibs object
    assert a == '0b11111010'
    a.prepend(Tibs.from_bools([True, False, True]))  # Boolean list
    assert a == '0b10111111010'

    # Empty prepend
    a = Mutibs('0x42')
    a.prepend(Tibs())
    assert a == '0x42'


def test_append_prepend_together():
    # Test combining both operations
    a = Mutibs('0xAA')
    a.append('0xBB').prepend('0xCC')
    assert a == '0xCCAABB'


def test_setitem_single_bit():
    a = Mutibs('0b0010')
    a[0] = 1
    assert a == '0b1010'
    a[2] = 0
    assert a == '0b1000'
    a[-1] = True
    assert a == '0b1001'
    a[-4] = False
    assert a == '0b0001'
    # Out of range
    with pytest.raises(IndexError):
        a[4] = 1
    with pytest.raises(IndexError):
        a[-5] = 0


def test_setitem_slice():
    a = Mutibs('0b101010')
    a[1:4] = '0b111'
    assert a == '0b111110'
    a[0:2] = Tibs('0b00')
    assert a == '0b001110'
    a[2:5] = Mutibs('0b101')
    assert a == '0b001010'
    # Negative indices
    a[-3:-1] = '0b11'
    assert a == '0b001110'
    # Full slice
    a[:] = '0b000000'
    assert a == '0b000000'
    # Empty slice
    a[2:2] = '0b'
    assert a == '0b000000'
    a[1:3] = '0b1'
    assert a == '0b01000'
    # Stepping is not allowed
    with pytest.raises(ValueError):
        a[::2] = '0b00'
    a[10:12] = '0b00'  # Out of range, so just extends.
    assert a == '0b0100000'


def test_setitem_slice_length_change():
    a = Mutibs('0b1010')
    a[1:3] = '0b111'
    assert a == '0b11110'  # Length increased by 1
    a[0:2] = '0b0'
    assert a == '0b0110'
    a[1:2] = '0b1111'
    assert a == '0b0111110'
    a[0:15] = '0b1'
    assert a == '0b1'
    # Setting to empty
    a[:] = ''
    assert a == ''
    # Setting empty slice to non-empty
    a[0:0] = '0b101'
    assert a == '0b101'


def test_delitem_single_bit():
    # Test deleting single bits
    a = Mutibs('0b1010')
    del a[1]
    assert a == '0b110'

    a = Mutibs('0b1010')
    del a[-1]
    assert a == '0b101'

    # Out of range
    with pytest.raises(IndexError):
        a = Mutibs('0b101')
        del a[3]

    with pytest.raises(IndexError):
        a = Mutibs('0b101')
        del a[-4]


def test_delitem_slice():
    # Test deleting slices
    a = Mutibs('0b101010')
    del a[1:4]
    assert a == '0b110'

    # Negative indices
    a = Mutibs('0b101010')
    del a[-4:-2]
    assert a == '0b1010'

    # Empty slice should do nothing
    a = Mutibs('0b1010')
    del a[2:2]
    assert a == '0b1010'

    # Full slice deletion
    a = Mutibs('0b1010')
    del a[:]
    assert a == ''

    # Partial indices
    a = Mutibs('0b101010')
    del a[2:]  # Delete from index 2 to the end
    assert a == '0b10'

    a = Mutibs('0b101010')
    del a[:2]  # Delete from start to index 2
    assert a == '0b1010'


def test_delitem_with_step():
    # Test slices with step
    a = Mutibs('0b101010')
    del a[::2]  # Delete every other bit
    assert a == '0b000'
    with pytest.raises(ValueError):
        del a[::0]


def test_delitem_edge_cases():
    # Empty bits
    a = Mutibs()
    with pytest.raises(IndexError):
        del a[0]

    a = Mutibs('0b1010')
    del a[10:20]  # Out of range slice, should do nothing
    assert a == '0b1010'

    # Delete last bit
    a = Mutibs('0b1')
    del a[0]
    assert a == ''


def test_inplace_add():
    a = Mutibs('0x123')
    a += '0xff'
    assert a == '0x123ff'


def test_inplace_mul():
    a = Mutibs()
    a *= 10000
    assert a == Mutibs()
    a += '0b10'
    a *= 5
    assert a == '0b1010101010'


def test_or():
    a = Mutibs('0x0f')
    b = Mutibs('0xf0')
    c = a | b
    assert c == '0xff'


def test_ior():
    a = Mutibs('0xf00')
    a |= '0x00a'
    assert a == '0xf0a'


def test_iand():
    a = Mutibs('0b1100')
    a &= '0b1010'
    assert a == '0b1000'
    b = Mutibs('0b1111')
    a &= b
    assert a == '0b1000'
    from tibs import Tibs
    c = Tibs('0b0100')
    a &= c
    assert a == '0b0000'


def test_and():
    a = Mutibs('0b1100')
    b = Mutibs('0b1010')
    c = a & b
    assert c == '0b1000'
    d = Tibs('0b0110')
    e = a & d
    assert e == '0b0100'


def test_ixor():
    a = Mutibs('0b1100')
    a ^= '0b1010'
    assert a == '0b0110'
    b = Mutibs('0b0011')
    a ^= b
    assert a == '0b0101'
    c = Tibs('0b1100')
    a ^= c
    assert a == '0b1001'


def test_xor():
    a = Mutibs('0b1100')
    b = Mutibs('0b1010')
    c = a ^ b
    assert c == '0b0110'
    d = Tibs('0b0110')
    e = a ^ d
    assert e == '0b1010'


def test_constructors():
    a = Mutibs.from_f(0.5, 32)
    b = Mutibs.from_bytes(b'123')
    c = Mutibs.from_bin('100')
    d = Mutibs.from_oct('7654')

    b.prepend(b)
    assert b == Mutibs.from_bytes(b'123123')

    c.append(d)
    assert c == '0o47654'
    d.append(d)
    assert d == '0o76547654'

def test_invert():
    a = Mutibs('0b1110')
    b = ~a
    assert b == '0b0001'
    assert a == '0b1110'


def test_insert_basic():
    # Basic insert functionality
    a = Mutibs('0b1010')
    a.insert(2, '0b11')
    assert a == '0b101110'


def test_insert_beginning():
    # Insert at beginning
    a = Mutibs('0b1010')
    a.insert(0, '0b11')
    assert a == '0b111010'


def test_insert_end():
    # Insert at end
    a = Mutibs('0b1010')
    a.insert(4, '0b11')
    assert a == '0b101011'


def test_insert_empty():
    # Insert empty bits
    a = Mutibs('0b1010')
    a.insert(2, '')
    assert a == '0b1010'


def test_insert_from_bits():
    # Insert with Tibs object
    a = Mutibs('0b1010')
    a.insert(2, Tibs('0b11'))
    assert a == '0b101110'


def test_insert_from_mutable_bits():
    # Insert with Mutibs object
    a = Mutibs('0b1010')
    a.insert(2, Mutibs('0b11'))
    assert a == '0b101110'


def test_insert_chaining():
    # Method chaining
    a = Mutibs('0b10')
    result = a.insert(1, '0b1').insert(2, '0b0')
    assert a == '0b1100'
    assert result is a


def test_insert_beyond_length():
    # Position beyond length
    a = Mutibs('0b1010')
    a.insert(5, '0b11')  # Position beyond length
    assert a == '0b101011'  # Just appends - standard Python behaviour


def test_set_single_bit_to_one():
    # Basic set functionality - setting a single bit to 1
    a = Mutibs('0b0000')
    a.set(1, 2)
    assert a == '0b0010'


def test_set_single_bit_to_zero():
    # Setting a single bit to 0
    a = Mutibs('0b1111')
    a.set(0, 2)
    assert a == '0b1101'


def test_set_with_boolean_values():
    # Setting with boolean values
    a = Mutibs('0b0000')
    a.set(True, 1)
    assert a == '0b0100'
    a.set(False, 1)
    assert a == '0b0000'


def test_set_with_negative_index():
    # Setting with negative index
    a = Mutibs('0b0010')
    a.set(1, -1)
    assert a == '0b0011'
    a.set(0, -2)
    assert a == '0b0001'


def test_set_multiple_positions():
    # Setting multiple positions
    a = Mutibs('0b0000')
    a.set(1, [0, 2])
    assert a == '0b1010'


def test_set_mixed_indices():
    # Setting with mixed positive and negative indices
    a = Mutibs('0b0000')
    a.set(1, [1, -1])
    assert a == '0b0101'


def test_set_with_range():
    # Setting with range
    a = Mutibs('0b0000')
    a.set(1, range(4))
    assert a == '0b1111'


def test_set_with_empty_sequence():
    # Setting with an empty sequence
    a = Mutibs('0b1010')
    a.set(0, [])
    assert a == '0b1010'  # Should remain unchanged


def test_set_method_chaining():
    # Method chaining
    a = Mutibs('0b0000')
    result = a.set(1, 0).set(1, 2)
    assert a == '0b1010'
    assert result is a


def test_set_with_non_boolean_values():
    # Testing non-boolean values
    a = Mutibs('0b0000')
    a.set("string", 1)  # Any non-empty string evaluates to True
    assert a == '0b0100'
    a.set(0, 1)  # 0 evaluates to False
    assert a == '0b0000'


def test_set_index_out_of_range():
    # Error cases
    with pytest.raises(IndexError):
        a = Mutibs('0b1010')
        a.set(1, 4)  # Index out of range


def test_set_negative_index_out_of_range():
    with pytest.raises(IndexError):
        a = Mutibs('0b1010')
        a.set(0, -5)  # Negative index out of range


def test_invert_all():
    # Test invert method with no argument (inverts all bits)
    a = Mutibs('0b1010')
    a.invert()
    assert a == '0b0101'


def test_invert_single_bit():
    # Test inverting single bit
    a = Mutibs('0b1010')
    a.invert(1)
    assert a == '0b1110'


def test_invert_with_negative_index():
    # Test with negative index
    a = Mutibs('0b1010')
    a.invert(-1)
    assert a == '0b1011'


def test_invert_multiple_positions():
    # Test with list of positions
    a = Mutibs('0b1010')
    a.invert([0, 2])
    assert a == '0b0000'


def test_invert_mixed_indices():
    # Test with mixed positive and negative indices
    a = Mutibs('0b1010')
    a.invert([0, -2])
    assert a == '0b0000'


def test_invert_with_range():
    # Test with range
    a = Mutibs('0b1010')
    a.invert(range(2))
    assert a == '0b0110'


def test_invert_chaining():
    # Method chaining
    a = Mutibs('0b1010')
    result = a.invert(1).invert(2)
    assert a == '0b1100'
    assert result is a


def test_invert_index_out_of_range():
    # Error cases
    with pytest.raises(IndexError):
        a = Mutibs('0b1010')
        a.invert(4)  # Index out of range


def test_invert_negative_index_out_of_range():
    with pytest.raises(IndexError):
        a = Mutibs('0b1010')
        a.invert(-5)  # Negative index out of range


def test_invert_empty_bits():
    # Empty Mutibs
    a = Mutibs()
    a.invert()  # Inverting empty bits should do nothing
    assert a == ''



def test_replace_basic():
    # Basic replace functionality
    a = Mutibs('0b10101010')
    a.replace('0b10', '0b111')
    assert a == '0b111111111111'

def test_replace_same_length():
    # Replace with same length pattern
    a = Mutibs('0b10101010')
    a.replace('0b10', '0b00')
    assert a == '0b00000000'

def test_replace_with_empty():
    # Replace with empty bits (should effectively delete)
    a = Mutibs('0b10101010')
    a.replace('0b10', '')
    assert a == ''

def test_replace_with_count():
    # Replace only first occurrences with count parameter
    a = Mutibs('0b10101010')
    a.replace('0b10', '0b00', count=2)
    assert a == '0b00001010'

def test_replace_with_start():
    # Replace with start parameter
    a = Mutibs('0b10101010')
    a.replace('0b10', '0b11', start=2)
    assert a == '0b10111111'

def test_replace_with_end():
    # Replace with end parameter
    a = Mutibs('0b10101010')
    a.replace('0b10', '0b11', end=4)
    assert a == '0b11111010'

def test_replace_with_start_end():
    # Replace with both start and end parameters
    a = Mutibs('0b10101010')
    a.replace('0b10', '0b11', start=2, end=6)
    assert a == '0b10111110'

def test_replace_byte_aligned():
    # Replace with byte_aligned=True
    a = Mutibs('0b10101010')
    a.replace('0b1010', '0b1111', byte_aligned=True)
    assert a == '0b11111010'

def test_replace_method_chaining():
    # Method chaining
    a = Mutibs('0b10101010')
    result = a.replace('0b10', '0b11').replace('0b11', '0b00')
    assert a == '0b00000000'
    assert result is a

def test_replace_different_types():
    # Replace with different types
    a = Mutibs('0b10101010')
    a.replace(Tibs('0b10'), Mutibs('0b11'))
    assert a == '0b11111111'

def test_replace_empty_pattern():
    # Empty pattern (should raise error)
    with pytest.raises(ValueError):
        a = Mutibs('0b1010')
        a.replace('', '0b11')

def test_replace_pattern_not_found():
    # Pattern not found
    a = Mutibs('0b1010')
    a.replace('0b11', '0b00')
    assert a == '0b1010'  # Should remain unchanged

def test_replace_with_count_zero():
    # Count=0 (should not replace anything)
    a = Mutibs('0b10101010')
    a.replace('0b10', '0b11', count=0)
    assert a == '0b10101010'

def test_reverse_basic():
    # Basic reverse functionality
    a = Mutibs('0b1010')
    a.reverse()
    assert a == '0b0101'


def test_reverse_palindrome():
    # Palindrome should remain the same when reversed
    a = Mutibs('0b1001')
    a.reverse()
    assert a == '0b1001'


def test_reverse_empty():
    # Reverse empty Mutibs
    a = Mutibs()
    a.reverse()
    assert a == ''


def test_reverse_single_bit():
    # Reverse single bit
    a = Mutibs('0b1')
    a.reverse()
    assert a == '0b1'


def test_reverse_hex():
    # Reverse with hex representation
    a = Mutibs('0xAB')
    a.reverse()
    assert a == '0xd5'  # 0xAB = 10101011 -> 11010101 = 0xd5


def test_reverse_method_chaining():
    # Method chaining
    a = Mutibs('0b1100')
    result = a.reverse()
    assert a == '0b0011'
    assert result is a


def test_reverse_idempotence():
    # Reverse twice should give original
    a = Mutibs('0b10110')
    a.reverse().reverse()
    assert a == '0b10110'


def test_rol_basic():
    # Basic rotate left functionality
    a = Mutibs('0b1010')
    a.rol(1)
    assert a == '0b0101'


def test_rol_full_rotation():
    # Rotating by the full length should return the original
    a = Mutibs('0b1010')
    a.rol(4)
    assert a == '0b1010'


def test_rol_wraparound():
    # Rotating by more than length should wrap around
    a = Mutibs('0b1010')
    a.rol(5)
    assert a == '0b0101'  # Same as rol(1)


def test_rol_with_start_end():
    # Rotating with start and end parameters
    a = Mutibs('0b10101100')
    a.rol(2, start=2, end=6)
    assert a == '0b10111000'


def test_rol_method_chaining():
    # Method chaining
    a = Mutibs('0b1010')
    result = a.rol(1)
    assert a == '0b0101'
    assert result is a


def test_rol_negative_amount():
    # Error cases - negative rotation
    with pytest.raises(ValueError):
        a = Mutibs('0b1010')
        a.rol(-1)  # Negative rotation amount


def test_rol_empty_bits():
    # Error cases - empty bits
    with pytest.raises(ValueError):
        a = Mutibs()
        a.rol(1)  # Empty Mutibs


def test_rol_zero_rotation():
    # Zero rotation should not change anything
    a = Mutibs('0b1010')
    a.rol(0)
    assert a == '0b1010'


def test_rol_large_rotation():
    # Large rotation value
    a = Mutibs('0b1010')
    a.rol(1000000)  # Should be equivalent to rol(0) since 1000000 % 4 = 0
    assert a == '0b1010'


def test_ror_basic():
    # Basic rotate right functionality
    a = Mutibs('0b1010')
    a.ror(1)
    assert a == '0b0101'


def test_ror_full_rotation():
    # Rotating by the full length should return the original
    a = Mutibs('0b1010')
    a.ror(4)
    assert a == '0b1010'


def test_ror_wraparound():
    # Rotating by more than length should wrap around
    a = Mutibs('0b1010')
    a.ror(5)
    assert a == '0b0101'  # Same as ror(1)


def test_ror_with_start_end():
    # Rotating with start and end parameters
    a = Mutibs('0b10101100')
    a.ror(2, start=2, end=6)
    assert a == '0b10111000'


def test_ror_method_chaining():
    # Method chaining
    a = Mutibs('0b1010')
    result = a.ror(1)
    assert a == '0b0101'
    assert result is a


def test_rol_ror_cancellation():
    # Rotating left then right should cancel out
    a = Mutibs('0b10110')
    a.rol(2).ror(2)
    assert a == '0b10110'


def test_ror_negative_amount():
    # Error cases - negative rotation
    with pytest.raises(ValueError):
        a = Mutibs('0b1010')
        a.ror(-1)  # Negative rotation amount


def test_ror_empty_bits():
    # Error cases - empty bits
    with pytest.raises(ValueError):
        a = Mutibs()
        a.ror(1)  # Empty Mutibs


def test_ror_zero_rotation():
    # Zero rotation should not change anything
    a = Mutibs('0b1010')
    a.ror(0)
    assert a == '0b1010'


def test_ror_large_rotation():
    # Large rotation value
    a = Mutibs('0b1010')
    a.ror(1000000)  # Should be equivalent to ror(0) since 1000000 % 4 = 0
    assert a == '0b1010'


def test_byte_swap_basic():
    # Basic byte_swap functionality with default parameters
    a = Mutibs('0x1234')
    a.byte_swap()
    assert a == '0x3412'


def test_byte_swap_with_length():
    # Byte swap with specific byte_length parameter
    a = Mutibs('0x12345678')
    a.byte_swap(2)
    assert a == '0x34127856'


def test_byte_swap_single_byte():
    # Byte swap single byte (no change)
    a = Mutibs('0x12')
    a.byte_swap(1)
    assert a == '0x12'


def test_byte_swap_method_chaining():
    # Method chaining
    a = Mutibs('0x1234')
    result = a.byte_swap()
    assert a == '0x3412'
    assert result is a


def test_byte_swap_idempotence():
    # Byte swap twice should return to original
    a = Mutibs('0x12345678')
    a.byte_swap(2).byte_swap(2)
    assert a == '0x12345678'


def test_byte_swap_non_multiple_of_8():
    # Non-multiple of 8 bits
    with pytest.raises(ValueError):
        a = Mutibs('0b10101')
        a.byte_swap()


def test_byte_swap_empty():
    # Empty Mutibs
    a = Mutibs()
    result = a.byte_swap()
    assert a == ''
    assert result == Mutibs()


def test_byte_swap_negative_length():
    # Negative byte length
    with pytest.raises(ValueError):
        a = Mutibs('0x1234')
        a.byte_swap(-1)


def test_byte_swap_zero_length():
    # Zero byte length
    with pytest.raises(ValueError):
        a = Mutibs('0x1234')
        a.byte_swap(0)


def test_byte_swap_not_multiple_of_byte_length():
    # Not a multiple of byte_length
    with pytest.raises(ValueError):
        a = Mutibs('0x123456')  # 3 bytes
        a.byte_swap(2)  # Not a multiple of 2 bytes


def test_to_tibs_basic():
    # Basic conversion
    a = Mutibs('0b1010')
    b = a.to_tibs()
    assert isinstance(b, Tibs)
    assert b == '0b1010'


def test_to_tibs_immutable_copy_operations():
    # Original shouldn't change when immutable copy is modified
    a = Mutibs('0b1010')
    b = a.to_tibs()
    c = ~b
    assert a == '0b1010'  # Original remains unchanged
    assert b == '0b1010'  # Original immutable copy unchanged
    assert c == '0b0101'  # New inverted copy


def test_to_tibs_original_modifications():
    # Changes to original shouldn't affect the immutable copy
    a = Mutibs('0b1010')
    b = a.to_tibs()
    a.invert()
    assert a == '0b0101'  # Original changed
    assert b == '0b1010'  # Immutable copy remains unchanged


def test_to_tibs_empty():
    # Empty Mutibs conversion
    a = Mutibs()
    b = a.to_tibs()
    assert isinstance(b, Tibs)
    assert b == ''
    assert len(b) == 0


def test_mutable_bits_from_bits():
    # Test creating Mutibs from Tibs object
    b = Tibs('0b1010')
    a = b.to_mutibs()
    assert a == '0b1010'
    assert isinstance(a, Mutibs)

    # Modification should not affect original
    a.invert()
    assert a == '0b0101'
    assert b == '0b1010'


def test_setitem_with_bits_object():
    # Test setting slices using Tibs objects
    a = Mutibs('0b1010')
    b = Tibs('0b11')
    a[1:3] = b
    assert a == '0b1110'


def test_iadd_with_bits():
    # Test in-place add with Tibs objects
    a = Mutibs('0x12')
    b = Tibs('0x34')
    a += b
    assert a == '0x1234'


def test_iadd_multiple_types():
    # Test in-place add with various types
    a = Mutibs('0b1010')
    a += '0b11'  # String
    a += Tibs('0b00')  # Tibs object
    a += Mutibs('0b111')  # Another Mutibs
    assert a == '0b10101100111'


def test_imul_repeats():
    # Test in-place multiply
    a = Mutibs('0b101')
    a *= 3
    assert a == '0b101101101'

    # Test with zero
    b = Mutibs('0b111')
    b *= 0
    assert b == ''


def test_delitem_sequence():
    # Test deleting multiple items in sequence
    a = Mutibs('0b10101010')
    del a[0]
    assert a == '0b0101010'
    del a[2]
    assert a == '0b011010'
    del a[-1]
    assert a == '0b01101'


def test_setitem_complex_cases():
    # Test setting a slice with different-length content
    a = Mutibs('0b1010')
    a[1:3] = '0b111'  # Replace 2 bits with 3 bits
    assert a == '0b11110'

    # Replace with empty content (effectively deleting)
    a[2:4] = ''
    assert a == '0b110'

    # Replace everything with shorter content
    a[:] = '0b1'
    assert a == '0b1'


def test_bit_operations_with_bits():
    # Testing bitwise AND with Tibs
    a = Mutibs('0b1100')
    b = Tibs('0b1010')
    a &= b
    assert a == '0b1000'

    # Testing bitwise OR with Tibs
    a = Mutibs('0b1100')
    b = Tibs('0b0011')
    a |= b
    assert a == '0b1111'

    # Testing bitwise XOR with Tibs
    a = Mutibs('0b1100')
    b = Tibs('0b1010')
    a ^= b
    assert a == '0b0110'


def test_equality_with_bits():
    # Test equality comparison with Tibs
    a = Mutibs('0b1010')
    b = Tibs('0b1010')
    assert a == b

    # Test after modification
    a[0] = 0
    assert a != b
    assert a == '0b0010'


def test_interleaved_operations():
    # Test a sequence of interleaved operations
    a = Mutibs('0b1010')
    a[1:3] = '0b00'
    a += '0b11'
    a.invert(0)
    del a[-1]
    assert a == '0b00001'

    # Chain multiple operations
    a = Mutibs('0b101')
    result = a.append('0b010').invert().reverse()
    assert result == a  # Verify chaining returns self
    assert a == '0b101010'  # 101 + 010 -> 101010 -> 010101 (invert) -> 010010 (reverse)


def test_mutable_bits_conversion_roundtrip():
    # Test round-trip conversion between Tibs and Mutibs
    orig = Tibs('0b10101100')
    mutable = orig.to_mutibs()
    mutable.invert(range(4))  # Modify some bits
    back_to_tibs = mutable.to_tibs()

    assert isinstance(back_to_tibs, Tibs)
    assert back_to_tibs == '0b01011100'
    assert orig == '0b10101100'  # Original should be unchanged


def test_inserting_bits_objects():
    # Test inserting Tibs objects at specific positions
    a = Mutibs('0b1010')
    b = Tibs('0b11')
    a.insert(2, b)
    assert a == '0b101110'

    # Insert at beginning
    c = Tibs('0b00')
    a.insert(0, c)
    assert a == '0b00101110'


def test_mixed_representation_operations():
    # Test operations with mixed representations (binary, hex)
    a = Mutibs('0b1010')
    a += '0x3A'
    assert a == '0b1010_0011_1010'

    a[4:8] = '0o7'
    assert a == '0b1010_111_1010'


def test_shifting_inplace():
    # Test in-place shifting operations
    a = Mutibs('0b001010')
    a <<= 2
    assert a == '0b101000'
    a >>= 3
    assert a == '0b000101'
    with pytest.raises(ValueError):
        a <<= -1
    with pytest.raises(ValueError):
        a >>= -1


def test_all_any():
    a = Mutibs('0x00')
    assert not a.any()
    assert not a.all()
    b = Mutibs('0xff')
    assert b.any()
    assert b.all()


def test_shifts():
    a = Mutibs.from_ones(5)
    a += '0b0'
    b = a << 1
    assert b == '0b111100'
    c = b >> 1
    assert c == '0b011110'


def test_str():
    a = Mutibs.from_ones(8)
    assert a.__str__() == '0xff'
    assert a.__repr__() == "Mutibs('0xff')"


def test_logical_op_misc():
    a = Mutibs('0xffff')
    b = Mutibs('0x000')
    try:
        _ = a & b
    except ValueError as e:
        assert "12" in str(e)
        assert "16" in str(e)


def test_auto_conversions():
    a = Mutibs()
    with pytest.raises(TypeError):
        _ = a + None
    with pytest.raises(TypeError):
        _ = a + True
    with pytest.raises(TypeError):
        _ = a + False
    with pytest.raises(TypeError):
        _ = a + 1
    b = a + '0x1'
    assert isinstance(b, Mutibs) and b == '0x1'
    b = a + b'123'
    assert isinstance(b, Mutibs) and b == b'123'
    b = a + [1, 0]
    assert isinstance(b, Mutibs) and b == '0b10'
    b = a + (1, 0, 'steve')
    assert isinstance(b, Mutibs) and b == '0b101'


def test_clear():
    a = Mutibs()
    a.clear()
    assert a == Mutibs()
    assert not a
    a += '0b1'
    assert a
    a.clear()
    assert not a
    assert a == Mutibs()


def test_reserve():
    a = Mutibs()
    assert a.capacity() == 0
    a.reserve(10)
    assert a.capacity() >= 10
    a += Mutibs.from_random(1000000)
    b4 = a.capacity()
    assert b4 >= 1000000
    a.clear()
    assert a.capacity() == b4


def test_insert_slice():
    a = Mutibs('0xff')
    a[0:0] = '0xab'
    assert a == '0xabff'
    a[0:0] = a
    assert a == '0xabffabff'


def test_del_ranges():
    a = Mutibs.from_zeros(10)
    del a[5:3]
    assert len(a) == 10


def test_set_item_with_step():
    a = Mutibs('0b000000')
    a[::2] = '0b110'
    assert a == '0b101000'


def test_iter():
    a = Mutibs('0b110')
    with pytest.raises(TypeError):
        _ = [bool(q) for q in a]



#!/usr/bin/env python

import pytest
from tibs import Tibs, Mutibs
import math
import copy


class TestNoPosAttribute:
    def test_replace(self):
        s = Mutibs.from_string("0b01")
        s = s.replace("0b1", "0b11")
        assert s == "0b011"

    def test_delete(self):
        s = Mutibs('0b000000001')
        del s[-1:]
        assert s == '0b00000000'

    def test_insert(self):
        s = Mutibs.from_string("0b00")
        s = s.insert(1, "0xf")
        assert s == "0b011110"

    def test_insert_self(self):
        b = Mutibs.from_string("0b10")
        b = b.insert(0, b)
        assert b == "0b1010"
        c = Mutibs.from_string("0x00ff")
        c = c.insert(8, c)
        assert c == "0x0000ffff"
        a = Mutibs.from_string("0b11100")
        a = a.insert(3, a)
        assert a == "0b1111110000"

    def test_overwrite(self):
        s = Mutibs.from_string("0b01110")
        s[1:4] = "0b000"
        assert s == "0b00000"

    def test_prepend(self):
        s = Tibs.from_zeros(1)
        t = Tibs.from_bools([1]) + s
        assert s == "0b0"
        assert t == "0b10"

    def test_rol(self):
        s = Mutibs("0b0001")
        t = s.rol(1)
        assert t == "0b0010"

    def test_ror(self):
        s = Tibs.from_string("0b1000")
        t = s.to_mutibs().ror(1)
        assert s == "0b1000"
        assert t == "0b0100"

    def test_set_item(self):
        s = Mutibs('0b000100')
        s[4:5] = '0xf'
        assert s == '0b000111110'
        s[0:1] = [1]
        assert s == '0b100111110'
        s[5:5] = Mutibs()
        assert s == '0b100111110'

    def test_adding_nonsense(self):
        a = Mutibs.from_bools([0])
        with pytest.raises(ValueError):
            a += '3'
        with pytest.raises(ValueError):
            a += 'se'
        with pytest.raises(ValueError):
            a += 'float:32'


def test_not_byte_aligned():
    a = Tibs.from_string("0x00 ff 0f f")
    li = list(a.find_all("0xff"))
    assert li == [8, 20]
    p = a.find("0x0f")
    assert p == 4
    p = a.rfind("0xff")
    assert p == 20
    a = a.to_mutibs().replace("0xff", "")
    assert a == "0x000"


class TestSliceAssignment:

    def test_slice_assignment_single_bit_errors(self):
        a = Mutibs('0b000')
        with pytest.raises(IndexError):
            a[-4] = 1
        with pytest.raises(IndexError):
            a[3] = 1

    def test_slice_assignment_multiple_bits(self):
        a = Mutibs('0b0')
        a[0:1] = '0b110'
        assert a.to_bin() == '110'
        a[0:1] = '0b000'
        assert a.to_bin() == '00010'
        a[0:3] = '0b111'
        assert a.to_bin() == '11110'
        a[-2:] = '0b011'
        assert a.to_bin() == '111011'
        a[:] = '0x12345'
        assert a.to_hex() == '12345'
        a[:] = ''
        assert not a

    def test_slice_assignment_multiple_bits_errors(self):
        a = Mutibs()
        with pytest.raises(IndexError):
            a[0] = '0b00'
        a += '0b1'
        a[0:2] = '0b11'
        assert a == '0b11'

    def test_del_slice_step(self):
        a = Mutibs.from_bin('100111101001001110110100101')
        del a[::2]
        assert a.to_bin() == '0110010101100'
        del a[3:9:3]
        assert a.to_bin() == '01101101100'
        del a[2:7:1]
        assert a.to_bin() == '011100'
        del a[::99]
        assert a.to_bin() == '11100'
        del a[::1]
        assert a.to_bin() == ''

    def test_del_slice_negative_step(self):
        a = Mutibs('0b0001011101101100100110000001')
        del a[5:23:-3]
        assert a.to_bin() == '0001011101101100100110000001'
        del a[25:3:-3]
        assert a.to_bin() == '00011101010000100001'
        del a[:6:-7]
        assert a.to_bin() == '000111010100010000'
        del a[15::-2]
        assert a.to_bin() == '0010000000'
        del a[::-1]
        assert a.to_bin() == ''

    def test_del_slice_negative_end(self):
        a = Mutibs('0b01001000100001')
        del a[:-5]
        assert a == '0b00001'
        a = Mutibs('0b01001000100001')
        del a[-11:-5]
        assert a == '0b01000001'

    def test_del_slice_errors(self):
        a = Mutibs.from_zeros(10)
        del a[5:3]
        assert a == Tibs.from_zeros(10)
        del a[3:5:-1]
        assert a == Tibs.from_zeros(10)

    def test_del_single_element(self):
        a = Mutibs('0b0010011')
        del a[-1]
        assert a.to_bin() == '001001'
        del a[2]
        assert a.to_bin() == '00001'
        with pytest.raises(IndexError):
            del a[5]

    def test_set_slice_step(self):
        a = Mutibs.from_bin('0000000000')
        a[::2] = '0b11111'
        assert a.to_bin() == '1010101010'
        a[4:9:3] = [0, 0]
        assert a.to_bin() == '1010001010'
        a[7:3:-1] = [1, 1, 1, 0]
        assert a.to_bin() == '1010011110'
        a[7:1:-2] = [0, 0, 1]
        assert a.to_bin() == '1011001010'
        a[::-5] = [1, 1]
        assert a.to_bin() == '1011101011'
        a[::-1] = [0, 0, 0, 0, 0, 0, 0, 0, 0, 1]
        assert a.to_bin() == '1000000000'

    def test_set_slice_step_with_int(self):
        a = Mutibs.from_zeros(9)
        with pytest.raises(TypeError):
            a[5:8] = -1

    def test_set_slice_errors(self):
        a = Mutibs.from_zeros(8)
        with pytest.raises(ValueError):
            a[::3] = [1]

        class A(object):
            pass

        with pytest.raises(TypeError):
            a[1:2] = A()
        with pytest.raises(ValueError):
            a[1:4:-1] = [1, 2]


def test_adding():
    a = Tibs.from_string("0b0")
    b = Tibs.from_string("0b11")
    c = a + b
    assert c == "0b011"
    assert a == "0b0"
    assert b == "0b11"


def test_copy_method():
    s = Tibs.from_zeros(9000)
    t = copy.copy(s)
    assert s == t
    assert s is t
    s = s.to_mutibs()
    t = copy.copy(s)
    assert s == t
    assert s is not t


class TestRepr:
    def test_standard_repr(self):
        a = Tibs.from_string("0o12345")
        assert repr(a).splitlines()[0] == "Tibs('0b001010011100101')"


def test_bits_conversion_to_bytes():
    a = Tibs.from_string("0x41424344, 0b1")
    with pytest.raises(ValueError):
        _ = bytes(a)
    b = bytes(a + [0, 0, 0, 0, 0, 0, 0])
    assert b == b"ABCD\x80"
    a = Tibs()
    assert bytes(a) == b""


def test_mutable_bits_conversion_to_bytes():
    a = Mutibs('0x0001')
    b = bytes(a)
    assert b == b'\x00\x01'
    b = bytes(Mutibs())
    assert b == b''


def test_bytes_from_list():
    s = Tibs.from_bytes(bytes([1, 2]))
    assert s == "0x0102"
    s = Tibs.from_bytes(bytearray([1, 2]))
    assert s == "0x0102"



# Comparison of performance between bitarray and tibs.
# This isn't meant as a competition, but more of a sanity check.
# If the tibs speed for a task is significantly less than the speed that bitarray can
# do the same task, then that points to an area that needs to be optimized.


import os
import sys

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import timeit
import random
import math
from math import isqrt
from random import randrange
from tibs import Tibs, Mutibs
from bitarray.util import random_p, ones
from bitarray import bitarray
from bitarray.util import int2ba, ba2int, pprint

ba_rand = random_p(1_000_000_000)
tibs_rand = Tibs.from_random(1_000_000_000)

some_bytes = Tibs.from_random(10_000_000, seed=b'a').to_bytes()
other_bytes = Tibs.from_random(10_000_000, seed=b'b').to_bytes()


def test_findall_tibs():
    t = Tibs.from_bytes(some_bytes)
    x = list(t.find_all('0xabc'))


def test_findall_bitarray():
    b = bitarray()
    b.frombytes(some_bytes)
    pattern = bitarray('101010111100')
    x = list(b.search(pattern))


def test_bitops_tibs():
    t1 = Tibs.from_bytes(some_bytes)
    t2 = Tibs.from_bytes(other_bytes)
    for _ in range(100):
        t3 = t1 | t2
        t4 = t3[10:1_000_000] & t2[9:999_999]
    print(t4.count(1))


def test_bitops_bitarray():
    b1 = bitarray()
    b1.frombytes(some_bytes)
    b2 = bitarray()
    b2.frombytes(other_bytes)
    for _ in range(100):
        b3 = b1 | b2
        b4 = b3[10:1_000_000] & b2[9:999_999]
    print(b4.count())


def test_construction_bitarray():
    b = bitarray()
    x = bitarray('10101')
    for _ in range(1000000):
        b.extend(x)
    assert len(b) == 5 * 1000000


def test_construction_tibs():
    t = Mutibs()
    x = Tibs('0b10101')
    for _ in range(1000000):
        t += x
    assert len(t) == 5 * 1000000


def test_counting_bitarray():
    for _ in range(100):
        _ = ba_rand.count(1)


def test_counting_tibs():
    for _ in range(100):
        _ = tibs_rand.count(1)


def test_rand_bitarray():
    s = random_p(1_000_000_000)


def test_rand_tibs():
    s = Mutibs.from_random(1_000_000_000)


def test_primes_bitarray():
    limit = 50_000_000
    is_prime = ones(limit)
    is_prime[:2] = False

    for i in range(2, isqrt(limit) + 1):
        if is_prime[i]:
            is_prime[i * i:: i] = False
    x = is_prime.count(bitarray("101")) + 1
    assert x == 239101


def test_primes_tibs():
    limit = 50_000_000
    is_prime = Mutibs.from_ones(limit)
    is_prime.set(False, [0, 1])
    for i in range(2, isqrt(limit) + 1):
        if is_prime[i]:
            is_prime.set(False, range(i * i, limit, i))
    twin_primes = len(list(is_prime.as_tibs().find_all('0b101')))
    assert twin_primes == 239101


class FunctionPairs:
    def __init__(self, name, bitarray_func, tibs_func):
        self.name = name
        self.bitarray_func = bitarray_func
        self.tibs_func = tibs_func
        self.bf_time = None
        self.bs_time = None
        self.ratio = 1.0

    def run(self):
        self.ba_time = timeit.timeit(self.bitarray_func, number=5)
        self.t_time = timeit.timeit(self.tibs_func, number=5)
        self.ratio = self.ba_time / self.t_time


class TestSuite:
    def __init__(self, pairs):
        self.pairs = pairs

    def run(self):
        for pair in self.pairs:
            pair.run()

    def print_results(self):
        for pair in self.pairs:
            if pair.ratio > 1.0:
                extra = ""
            else:
                extra = f"({1 / pair.ratio:.2f}⨉ slower)"
            print(
                f'{pair.name}: {pair.ratio:.2f}⨉ faster {extra} bitarray: {pair.ba_time:.2f}s vs tibs: {pair.t_time:.2f}s')
        # For ratios we use a geometric mean
        average = math.prod(r.ratio for r in self.pairs) ** (1 / len(self.pairs))
        print(f"AVERAGE: {average:.2f}⨉ faster")


def main():
    fn_pairs = [

        FunctionPairs("Primes", test_primes_bitarray, test_primes_tibs),
        FunctionPairs("Counting", test_counting_bitarray, test_counting_tibs),
        FunctionPairs("Random Generation", test_rand_bitarray, test_rand_tibs),
        FunctionPairs("Construction", test_construction_bitarray, test_construction_tibs),
        FunctionPairs("Find all", test_findall_bitarray, test_findall_tibs),
        FunctionPairs("Bit ops", test_bitops_bitarray, test_bitops_tibs),
    ]
    ts = TestSuite(fn_pairs)
    ts.run()
    ts.print_results()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import pytest
import io
import re
from hypothesis import given
import tibs
from tibs import Tibs, Mutibs
from typing import Iterable, Sequence


def test_from_bin():
    a = Tibs.from_bin('010')
    b = Tibs.from_string('0b010')
    c = Mutibs.from_bin('0b010')
    d = Tibs('0b010')
    assert a == b == c == d


def test_to_bin():
    a = Tibs('0b1001')
    assert a.to_bin() == '1001'
    assert a.to_mutibs().to_bin() == '1001'


def test_from_oct():
    a = Tibs.from_oct('12')
    b = Tibs.from_string('0o12')
    c = Mutibs.from_oct('0o12')
    d = Tibs('0o12')
    assert a == b == c == d


def test_to_oct():
    a = Tibs('0b001100')
    assert a.to_oct() == '14'
    assert a.to_mutibs().to_oct() == '14'


def test_from_hex():
    a = Tibs.from_hex('A')
    b = Tibs.from_string('0xA')
    c = Mutibs.from_hex('0xA')
    d = Tibs('0xA')
    assert a == b == c == d


def test_to_hex():
    a = Tibs('0b1010')
    assert a.to_hex() == 'a'
    assert a.to_mutibs().to_hex() == 'a'


def test_rfind():
    a = Mutibs()
    a += '0b1110001110'
    b = a.rfind('0b111')
    assert b == 6


def test_count_large():
    a = Tibs('0b' + '1' * 72)
    b = a[:65]
    assert b.count(1) == 65


def test_from_u():
    a = Tibs.from_u(15, 8)
    assert a == '0b00001111'
    b = Mutibs.from_u(15, 8)
    assert a == b
    c = a.to_u()
    assert c == 15


def test_from_u_errors():
    with pytest.raises(OverflowError):
        _ = Tibs.from_u(0, -1)
    with pytest.raises(OverflowError):
        _ = Tibs.from_u(0, 0)
    with pytest.raises(OverflowError):
        _ = Tibs.from_u(-1, 5)


def test_from_i():
    a = Tibs.from_i(-9, 100)
    b = Mutibs.from_i(-9, 100)
    assert a == b
    assert len(a) == 100
    assert a.to_i() == -9
    assert b.to_i() == -9


def test_from_i_errors():
    with pytest.raises(OverflowError):
        _ = Tibs.from_i(4, 2)


def test_from_large_ints():
    with pytest.raises(OverflowError):
        _ = Tibs.from_i(-1, 1000)
    a = Tibs.from_i(-1, 128)
    assert a.all()
    with pytest.raises(OverflowError):
        _ = Mutibs.from_i(-1, 1000)
    b = Mutibs.from_u(0, 128)
    assert not b.any()
    assert a.to_i() == -1
    assert b.to_u() == 0


def test_from_f():
    a = Tibs.from_f(0.25, 16)
    b = Tibs.from_f(0.25, 32)
    c = Tibs.from_f(0.25, 64)
    a2 = Mutibs.from_f(0.25, 16)
    b2 = Mutibs.from_f(0.25, 32)
    c2 = Mutibs.from_f(0.25, 64)
    assert a == a2
    assert len(a) == 16
    assert len(b) == 32
    assert len(c) == 64
    assert b == b2
    assert c == c2
    f = a.to_f()
    g = b.to_f()
    h = c.to_f()
    f2 = a2.to_f()
    g2 = b2.to_f()
    h2 = c2.to_f()
    assert f == g == h == f2 == g2 == h2 == 0.25



# Configuration file for the Sphinx documentation builder.
#
import os
import time
import datetime

year = datetime.datetime.utcfromtimestamp(
    int(os.environ.get("SOURCE_DATE_EPOCH", time.time()))
).year

project = "tibs"
copyright = f"2025 - {year}, Scott Griffiths"
author = "Scott Griffiths"
release = "0.3"

extensions = [
    "sphinx.ext.autodoc",
    'enum_tools.autoenum',
    'sphinx_autodoc_typehints',
]
autoapi_dirs = ["../tibs/"]
autoapi_add_toctree_entry = False

add_module_names = False

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

root_doc = "index"

add_function_parentheses = False

html_show_sphinx = False
html_static_path = ["_static"]
html_css_files = ["custom.css"]

html_sidebars = {
    "intro": []
}

html_theme = "pydata_sphinx_theme"
html_logo = "tibs.png"
html_theme_options = {
    "content_footer_items": ["last-updated"],
    "show_toc_level": 2,
    "logo": {
        # "text": "My awesome documentation",
        "image_light": "tibs.png",
        "image_dark": "tibs.png",
    },
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/scott-griffiths/tibs",
            "icon": "fa-brands fa-github",
            "type": "fontawesome",
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/tibs/",
            "icon": "fa-brands fa-python",
            "type": "fontawesome",
        },
    ],
    "footer_start": ["copyright"],
    "footer_end": ["last-updated"],
    "secondary_sidebar_items": ["page-toc"],
}

from pathlib import Path
from bs4 import BeautifulSoup


def process_in_page_toc(app, exception):
    for pagename in app.env.found_docs:
        if not isinstance(pagename, str):
            continue

        with (Path(app.outdir) / f"{pagename}.html").open("r") as f:
            # Parse HTML using BeautifulSoup html parser
            soup = BeautifulSoup(f.read(), "html.parser")

            for li in soup.find_all("li", class_="toc-h3 nav-item toc-entry"):
                if span := li.find("span"):
                    # Modify the toc-nav span element here
                    span.string = span.string.split(".")[-1]

        with (Path(app.outdir) / f"{pagename}.html").open("w") as f:
            # Write back HTML
            f.write(str(soup))


def setup(app):
    app.connect("build-finished", process_in_page_toc)



import sys
import pytest

sys.path.insert(0, "..")
from tibs import Tibs, Mutibs
import random
import math
import itertools


def test_chunking(benchmark):
    def chunks():
        s = Tibs.from_string("0xef1356a6200b3, 0b0")
        s = Tibs.from_joined(itertools.repeat(s, 6000))
        c = 0
        for triplet in s.chunks(3):
            if triplet == "0b001":
                c += 1
        return c

    c = benchmark(chunks)
    assert c == 12000, c


def test_count(benchmark):
    def count():
        s = Mutibs.from_zeros(100000000)
        s = s.set(1, [10, 100, 1000, 10000000])
        return s.count(1)

    c = benchmark(count)
    assert c == 4


def test_token_parsing(benchmark):
    def token_parsing():
        s = Mutibs()
        for i in range(10000):
            s += "0x3e44f, 0b11011, 0o75523"
            s += Tibs.from_bools([0, 1, 2, 0, 0, 1, 2, 0, -1, 0, "hello"])
            s += Tibs.from_zeros(104)

    benchmark(token_parsing)


def test_find_all(benchmark):
    def finding():
        s = Tibs.from_random(20000000, seed=b"99")
        for ss in [
            "0b11010010101",
            "0xabcdef1234, 0b000101111010101010011010100100101010101",
            "0x4321"
        ]:
            x = len(list(s.find_all(ss)))
        return x

    c = benchmark(finding)
    assert c == 305


def test_primes(benchmark):
    def primes():
        limit = 1000000
        is_prime = Mutibs.from_ones(limit)
        # Manually set 0 and 1 to be not prime.
        is_prime.set(False, [0, 1])
        # For every other integer, if it's set as prime then unset all of its multiples
        for i in range(2, math.ceil(math.sqrt(limit))):
            if is_prime[i]:
                is_prime.set(False, range(i * i, limit, i))
        twin_primes = len(list(is_prime.to_tibs().find_all("0b101")))
        return twin_primes

    c = benchmark(primes)
    assert c == 8169



from tibs import Tibs

start_code = '0x000001b3'
width = 352
height = 288
bit_rate = 5040
flags = [1, 1, 0]

t = Tibs.from_joined([
    start_code,
    Tibs.from_u(width, 12)
    Tibs.from_u(height, 12)
    Tibs.from_u(bit_rate, 13)
    flags])

assert t[0:32].to_hex() == '000001b3'
assert t[32:44].to_u() == width
assert t[-3:] == flags


#!/usr/bin/env python

import pytest
import copy
from tibs import Tibs, Mutibs


class TestFind:
    def test_find1(self):
        s = Tibs("0b0000110110000")
        assert s.find(Tibs("0b11011")) == 4

    def test_find_with_offset(self):
        s = Tibs("0x112233")[4:]
        assert s.find("0x23") == 8

    def test_find_corner_cases(self):
        s = Tibs("0b000111000111")
        assert s.find("0b000") == 0
        assert s.find("0b0111000111") == 2
        assert s[2:].find("0b000") == 4

    def test_find_bytes(self):
        s = Tibs.from_string("0x010203040102ff")
        assert s.find("0x05", byte_aligned=True) is None
        assert s.find("0x02", byte_aligned=True) == 8
        assert s[16:].find("0x02", byte_aligned=True) == 24
        assert s[1:].find("0x02", byte_aligned=True) == 0

    def test_find_bytes_aligned_corner_cases(self):
        s = Tibs("0xff")
        assert s.find(s) is not None
        assert s.find(Tibs("0x12")) is None
        assert s.find(Tibs("0xffff")) is None

    def test_find_byte_aligned(self):
        s = Tibs.from_hex("0x12345678")
        assert s.find(Tibs("0x56"), byte_aligned=True) == 16
        assert not s[16:].find(Tibs("0x45"), byte_aligned=True)
        s = Tibs("0x1234")
        assert s.find("0x1234") == 0

    def test_find_byte_aligned_with_offset(self):
        s = Tibs("0x112233")[4:]
        assert s.find("0x23", byte_aligned=True) == 8

    def test_find_byte_aligned_errors(self):
        s = Tibs("0xffff")
        with pytest.raises(ValueError):
            s.find("")
        with pytest.raises(ValueError):
            s.find(Tibs())


class TestRfind:
    def test_rfind(self):
        a = Tibs("0b001001001")
        b = a.rfind("0b001")
        assert b == 6
        big = Tibs.from_zeros(100000) + "0x12" + Tibs.from_zeros(10000)
        found = big.rfind("0x12", byte_aligned=True)
        assert found == 100000

    def test_rfind_byte_aligned(self):
        a = Tibs("0x8888")
        b = a.rfind("0b1", byte_aligned=True)
        assert b == 8

    def test_rfind_startbit(self):
        a = Tibs("0x0000ffffff")
        b = a.rfind("0x0000", start=1, byte_aligned=True)
        assert b is None

    def test_rfind_endbit(self):
        a = Tibs("0x000fff")
        b = a.rfind("0b011", start=0, end=14, byte_aligned=False)
        assert b is not None
        b = a.rfind("0b011", start=0, end=13, byte_aligned=False)
        assert b is None

    def test_rfind_errors(self):
        a = Tibs("0x43234234")
        with pytest.raises(ValueError):
            a.rfind("", byte_aligned=True)


class TestShift:
    def test_shift_left(self):
        s = Tibs.from_string("0b1010")
        t = s << 1
        assert s.to_bin() == "1010"
        assert t.to_bin() == "0100"
        s = t << 0
        assert s == "0b0100"
        t = s << 100
        assert t.to_bin() == "0000"

    def test_shift_left_errors(self):
        s = Tibs()
        with pytest.raises(ValueError):
            _ = s << 1
        s = Tibs("0xf")
        with pytest.raises(ValueError):
            _ = s << -1

    def test_shift_right(self):
        s = Tibs("0b1010")
        t = s >> 1
        assert s.to_bin() == "1010"
        assert t.to_bin() == "0101"
        q = s >> 0
        assert q == "0b1010"
        t = s >> 100
        assert t.to_bin() == "0000"

    def test_shift_right_errors(self):
        s = Tibs()
        with pytest.raises(ValueError):
            _ = s >> 1
        s = Tibs("0xf")
        with pytest.raises(ValueError):
            _ = s >> -1


class TestReplace:
    def test_replace1(self):
        a = Mutibs("0b1")
        a.replace("0b1", "0b0", byte_aligned=True)
        assert a.to_bin() == "0"
        a.replace("0b1", "0b0", byte_aligned=True)
        assert a.to_bin() == "0"

    def test_replace2(self):
        a = Mutibs("0b00001111111")
        a.replace("0b1", "0b0", byte_aligned=True)
        assert a.to_bin() == "00001111011"
        a.replace("0b1", "0b0", byte_aligned=False)
        assert a.to_bin() == "00000000000"

    def test_replace3(self):
        a = Mutibs("0b0")
        a.replace("0b0", "0b110011111", byte_aligned=True)
        assert a.to_bin() == "110011111"
        a.replace("0b11", "", byte_aligned=False)
        assert a.to_bin() == "001"

    def test_replace4(self):
        a = Mutibs("0x00114723ef4732344700")
        a.replace("0x47", "0x00", byte_aligned=True)
        assert a.to_hex() == "00110023ef0032340000"
        a.replace("0x00", "", byte_aligned=True)
        assert a.to_hex() == "1123ef3234"
        a.replace("0x11", "", start=1, byte_aligned=True)
        assert a.to_hex() == "1123ef3234"
        a.replace("0x11", "0xfff", start=7, byte_aligned=True)
        assert a.to_hex() == "1123ef3234"
        a.replace("0x11", "0xfff", start=0, byte_aligned=True)
        assert a.to_hex() == "fff23ef3234"

    def test_replace5(self):
        a = Tibs.from_string("0xab")
        b = Tibs.from_string("0xcd")
        c = Tibs.from_string("0xabef")
        c = c.to_mutibs().replace(a, b)
        assert c == "0xcdef"
        assert a == "0xab"
        assert b == "0xcd"
        a = Mutibs("0x0011223344").replace("0x11", "0xfff", byte_aligned=True)
        assert a == "0x00fff223344"

    def test_replace_with_self(self):
        a = Mutibs("0b11")
        a.replace("0b1", a)
        assert a == "0xf"
        a.replace(a, a)
        assert a == "0xf"

    def test_replace_count(self):
        a = Mutibs("0x223344223344223344")
        a.replace("0x2", "0x0", count=0, byte_aligned=True)
        assert a.to_hex() == "223344223344223344"
        a.replace("0x2", "0x0", count=1, byte_aligned=True)
        assert a.to_hex() == "023344223344223344"
        a.replace("0x33", "", count=2, byte_aligned=True)
        assert a.to_hex() == "02442244223344"
        a.replace("0x44", "0x4444", count=1435, byte_aligned=True)
        assert a.to_hex() == "02444422444422334444"

    def test_replace_errors(self):
        a = Mutibs("0o123415")
        with pytest.raises(ValueError):
            a.replace("", Tibs("0o7"), byte_aligned=True)


def test_empty_bitstring():
    s = Tibs()
    assert s.to_bin() == ""
    assert s.to_hex() == ""
    assert not s


class TestAppend:
    def test_append(self):
        s1 = Tibs("0b00000")
        s1 = s1 + Tibs.from_bools([1])
        assert s1.to_bin() == "000001"
        assert (Tibs("0x0102") + Tibs("0x0304")).to_hex() == "01020304"

    def test_append_same_bitstring(self):
        s1 = Tibs("0xf0")[:6]
        s1 = s1 + s1
        assert s1.to_bin() == "111100111100"


def test_insert():
    s = Mutibs("0x0011")
    s = s.insert(8, Tibs("0x22"))
    assert s.to_hex() == "002211"
    s = Mutibs.from_ones(0)
    s = s.insert(0, "0b101")
    assert s.to_bin() == "101"


class TestSlice:
    def test_byte_aligned_slice(self):
        s = Tibs("0x123456")
        assert s[8:16].to_hex() == "34"
        s = s[8:24]
        assert len(s) == 16
        assert s.to_hex() == "3456"
        s = s[0:8]
        assert s.to_hex() == "34"

    def test_slice(self):
        s = Tibs("0b000001111100000")
        s1 = s[0:5]
        s2 = s[5:10]
        s3 = s[10:15]
        assert s1.to_bin() == "00000"
        assert s2.to_bin() == "11111"
        assert s3.to_bin() == "00000"


class TestInsert:
    def test_insert(self):
        s1 = Mutibs("0x123456")
        s2 = Tibs("0xff")
        s1.insert(8, s2)
        assert s1.to_hex() == "12ff3456"
        s1.insert(24, "0xee")
        assert s1.to_hex() == "12ff34ee56"
        s1.insert(-1000, "0b1")  # Copying standard behavior of insert on lists
        assert s1 == '0b1, 0x12ff34ee56'
        s1.insert(1000, "0b1")  # Copying standard behavior of insert on lists
        assert s1 == '0b1, 0x12ff34ee56, 0b1'

    def test_insert_null(self):
        s = Mutibs("0x123")
        s.insert(3, Tibs())
        assert s.to_hex() == "123"

    def test_insert_bits(self):
        one = Tibs("0b1")
        zero = Tibs("0b0")
        s = Mutibs("0b00").insert(0, one)
        assert s.to_bin() == "100"
        s.insert(0, zero)
        assert s.to_bin() == "0100"
        s.insert(len(s), one)
        assert s.to_bin() == "01001"
        s.insert(2, s)
        assert s.to_bin() == "0101001001"


class TestOverwriting:
    def test_overwrite_bit(self):
        s = Tibs("0b0").to_mutibs()
        s[0:1] = "0b1"
        assert s.to_bin() == "1"

    def test_overwrite_limits(self):
        s = Mutibs.from_bin("0b11111")
        s[0:3] = "0b000"
        assert s.to_bin() == "00011"
        s[2:5] = "0b000"
        assert s.to_bin() == "00000"

    def test_overwrite_null(self):
        s = Mutibs("0x342563fedec")
        s2 = s[:]
        s[23:23] = Tibs()
        assert s.to_bin() == s2.to_bin()

    def test_overwrite_position(self):
        s1 = Mutibs("0x0123456")
        s2 = Tibs("0xff")
        s1[8: 8 + len(s2)] = s2
        assert s1.to_hex() == "01ff456"
        s1[0:8] = "0xff"
        assert s1.to_hex() == "ffff456"

    def test_overwrite_with_self(self):
        s = Mutibs("0x123")
        s[0:len(s)] = s
        assert s == "0x123"


class TestAdding:
    def test_adding(self):
        s1 = Tibs("0x0102")
        s2 = Tibs("0x0304")
        s3 = s1 + s2
        assert s1.to_hex() == "0102"
        assert s2.to_hex() == "0304"
        assert s3.to_hex() == "01020304"
        s3 += s1
        assert s3.to_hex() == "010203040102"
        assert s2[9:16].to_bin() == "0000100"
        assert s1[0:9].to_bin() == "000000010"
        s4 = Tibs.from_bin("000000010") + Tibs("0b0000100")
        assert s4.to_bin() == "0000000100000100"
        s5 = s1[0:9] + s2[9:16]
        assert s5.to_bin() == "0000000100000100"

    def test_more_adding(self):
        s = Tibs("0b00") + Tibs() + Tibs("0b11")
        assert s.to_bin() == "0011"
        s = "0b01"
        s += Tibs("0b11")
        assert s.to_bin() == "0111"
        s = Tibs("0x00")
        t = Tibs("0x11")
        s += t
        assert s.to_hex() == "0011"
        assert t.to_hex() == "11"
        s += s
        assert s.to_hex() == "00110011"

    def test_radd(self):
        s = "0xff" + Tibs("0xee")
        assert s.to_hex() == "ffee"

    def test_overwrite_more(self):
        s = Mutibs("0b11111")
        s[5:6] = "0b0"
        assert s.to_bin() == "111110"
        s[1:] = Tibs("0x00")
        assert s.to_bin() == "100000000"

    def test_get_item_with_positive_position(self):
        s = Tibs("0b1011")
        assert s[0] is True
        assert s[1] is False
        assert s[2] is True
        assert s[3] is True
        with pytest.raises(IndexError):
            _ = s[4]

    def test_get_item_with_negative_position(self):
        s = Tibs("0b1011")
        assert s[-1] is True
        assert s[-2] is True
        assert s[-3] is False
        assert s[-4] is True
        with pytest.raises(IndexError):
            _ = s[-5]

    def test_slicing(self):
        s = Tibs("0x0123456789")
        assert s[0:8].to_hex() == "01"
        assert not s[0:0]
        assert not s[23:20]
        assert s[8:12].to_bin() == "0010"
        assert s[32:80] == "0x89"

    def test_negative_slicing(self):
        s = Tibs("0x012345678")
        assert s[:-8].to_hex() == "0123456"
        assert s[-16:-8].to_hex() == "56"
        assert s[-24:].to_hex() == "345678"
        assert s[-1000:-24] == "0x012"

    def test_len(self):
        s = Tibs()
        assert len(s) == 0
        s = s + "0b001"
        assert len(s) == 3

    def test_join(self):
        s1 = Tibs("0b0")
        s2 = Tibs("0b1")
        s3 = Tibs("0b000")
        s4 = Tibs("0b111")
        strings = [s1, s2, s1, s3, s4]
        s = Tibs.from_joined(strings)
        assert s.to_bin() == "010000111"

    def test_join2(self):
        s1 = Tibs("0x00f1")
        assert s1[4:12].to_hex() == "0f"
        bsl = [s1[0:16], s1[4:12]]
        s = Tibs.from_joined(bsl)
        assert s.to_hex() == "00f10f"

        s1 = Tibs("0x00112233445566778899aabbccddeeff")
        s2 = Tibs("0b000011")
        bsl = [s1[0:32], s1[4:12], s2, s2, s2, s2]
        s = Tibs.from_joined(bsl)
        assert s.to_hex() == "00112233010c30c3"

    def test_join_with_ints(self):
        with pytest.raises(TypeError):
            _ = Tibs.from_joined([1, 2])

    def test_various_things2(self):
        s1 = Tibs("0x1f08")[:13]
        assert s1.to_bin() == "0001111100001"
        s2 = Tibs("0b0101")
        assert s2.to_bin() == "0101"
        s1 += s2
        assert len(s1) == 17
        assert s1.to_bin() == "00011111000010101"
        s1 = s1[3:8]
        assert s1.to_bin() == "11111"

    def test_various_things3(self):
        s1 = Tibs("0x012480ff")[2:27]
        s2 = s1 + s1
        assert len(s2) == 50
        s3 = s2[0:25]
        s4 = s2[25:50]
        assert s3.to_bin() == s4.to_bin()

    def test_insert_using_auto(self):
        s = Mutibs("0xff")
        s = s.insert(4, "0x00")
        assert s.to_hex() == "f00f"

    def test_overwrite_using_auto(self):
        s = Mutibs("0x0110")
        s[0:1] = "0b1"
        assert s.to_hex() == "8110"
        s[0:0] = ""
        assert s.to_hex() == "8110"

    def test_find_using_auto(self):
        s = Tibs("0b000000010100011000")
        assert s.find("0b101") == 7

    def test_findbyte_aligned_using_auto(self):
        s = Tibs("0x00004700")
        assert s.find("0b01000111", byte_aligned=True) == 16

    def test_append_using_auto(self):
        s = Tibs("0b000")
        s = s + "0b111"
        assert s.to_bin() == "000111"
        s = s + "0b0"
        assert s.to_bin() == "0001110"

    def test_prepend(self):
        s = Tibs("0b000")
        s = "0b11" + s
        assert s.to_bin() == "11000"
        s = s + s
        assert s.to_bin() == "1100011000"
        s = "" + s
        assert s.to_bin() == "1100011000"

    def test_null_slice(self):
        s = Tibs("0x111")
        t = s[1:1]
        assert len(t) == 0

    def test_multiple_autos(self):
        s = Tibs("0xa")
        s = "0xf" + s
        s = s + "0xb"
        assert s == "0xfab"
        s = s + s
        s = s + "0x100"
        print(type(s))
        with pytest.raises(TypeError):
            s[4: 8] = "0x5"
        s = s.to_mutibs()
        s[4:8] = "0x5"
        assert s == "0xf5bfab100"

    def test_reverse(self):
        s = Tibs("0b0011")
        s = s.to_mutibs().reverse()
        assert s.to_bin() == "1100"
        s = Mutibs("0b10")
        s.reverse()
        assert s.to_bin() == "01"
        s = Mutibs()
        s.reverse()
        assert s.to_bin() == ""

    def test_init_with_concatenated_strings(self):
        s = Tibs("0xff, 0xee,0xd ,0xcc")
        assert s.to_hex() == "ffeedcc"
        s = Tibs("0b0 ,0b111 ,0b001")
        assert s.to_bin() == "0111001"
        s = Tibs("0xffee")
        assert s.to_hex() == "ffee"
        s = Tibs("  0o123 ,0o7 ,0o1")
        assert s.to_oct() == "12371"
        s += "  0o 332"
        assert s.to_oct() == "12371332"

    def test_equals(self):
        s1 = Tibs("0b01010101")
        s2 = Tibs("0b01010101")
        assert s1 == s2
        s3 = Tibs()
        s4 = Tibs()
        assert s3 == s4
        assert not s3 != s4

    def test_large_equals(self):
        s1 = Tibs.from_zeros(1000000)
        s2 = Mutibs.from_zeros(1000000)
        s1 = s1.to_mutibs().set(True, [-1, 55, 53214, 534211, 999999])
        s2.set(True, [-1, 55, 53214, 534211, 999999])
        assert s1 == s2
        s1 = s1.set(True, 800000)
        assert s1 != s2

    def test_not_equals(self):
        s1 = Tibs("0b0")
        s2 = Tibs("0b1")
        assert s1 != s2
        assert not s1 != Tibs("0b0")

    def test_equality_with_auto_initialised(self):
        a = Tibs("0b00110111")
        assert a == "0b00110111"
        assert a == "0x37"
        assert "0b0011 0111" == a
        assert "0x37" == a
        assert not a == "0b11001000"
        assert not "0x3737" == a

    def test_invert_special_method(self):
        s = Tibs("0b00011001")
        assert (~s).to_bin() == "11100110"
        assert (~Tibs("0b0")).to_bin() == "1"
        assert (~Tibs("0b1")).to_bin() == "0"
        assert ~~s == s

    def test_invert_special_method_errors(self):
        s = Tibs()
        with pytest.raises(ValueError):
            _ = ~s

    def test_join_with_auto(self):
        s = Tibs.from_joined(["0xf", "0b00", Tibs.from_bin("11")])
        assert s == "0b11110011"


class TestMultiplication:
    def test_multiplication(self):
        a = Tibs("0xff")
        b = a * 8
        assert b == "0xffffffffffffffff"
        b = 4 * a
        assert b == "0xffffffff"
        assert 1 * a == a * 1 == a
        c = a * 0
        assert not c
        a *= 3
        assert a == "0xffffff"
        a *= 0
        assert not a
        one = Tibs("0b1")
        zero = Tibs("0b0")
        mix = one * 2 + 3 * zero + 2 * one * 2
        assert mix == "0b110001111"
        q = Tibs()
        q *= 143
        assert not q
        q += Tibs.from_bools([True, True, False])
        q *= 0
        assert not q

    def test_multiplication_errors(self):
        a = Tibs("0b1")
        b = Tibs("0b0")
        with pytest.raises(ValueError):
            _ = a * -1
        with pytest.raises(ValueError):
            a *= -1
        with pytest.raises(ValueError):
            _ = -1 * a
        with pytest.raises(TypeError):
            _ = a * 1.2
        with pytest.raises(TypeError):
            _ = b * a
        with pytest.raises(TypeError):
            a *= b


class TestBitWise:
    def test_bitwise_and(self):
        a = Tibs("0b01101")
        b = Tibs("0b00110")
        assert (a & b).to_bin() == "00100"
        assert (a & "0b11111") == a
        with pytest.raises(ValueError):
            _ = a & "0b1"
        with pytest.raises(ValueError):
            _ = b & "0b110111111"
        c = Tibs("0b0011011")
        d = c & "0b1111000"
        assert d.to_bin() == "0011000"
        d = "0b1111000" & c
        assert d.to_bin() == "0011000"

    def test_bitwise_or(self):
        a = Tibs("0b111001001")
        b = Tibs("0b011100011")
        c = a | b
        assert c.to_bin() == "111101011"
        assert (a | "0b000000000") == a
        with pytest.raises(ValueError):
            _ = a | "0b0000"
        with pytest.raises(ValueError):
            _ = b | (a + "0b1")
        a = "0xff00" | Tibs("0x00f0")
        assert a.to_hex() == "fff0"

    def test_bitwise_xor(self):
        a = Tibs("0b111001001")
        b = Tibs("0b011100011")
        c = a ^ b
        assert c.to_bin() == "100101010"
        assert (a ^ "0b111100000").to_bin() == "000101001"
        with pytest.raises(ValueError):
            _ = a ^ "0b0000"
        with pytest.raises(ValueError):
            _ = b ^ (a + "0b1")
        a = "0o707" ^ Tibs("0o777")
        assert a.to_oct() == "070"


def test_mutable_bitwise():
    a = '0xf0' & Mutibs('0x0f')
    assert a == '0x00'
    a = '0xf0' | Mutibs('0x0f')
    assert a == '0xff'
    a = '0xf0' & Mutibs('0x0f')
    assert a == '0x00'


class TestManyDifferentThings:
    def test_find_byte_aligned_with_bits(self):
        a = Tibs("0x00112233445566778899")
        x = a.find("0b0001", byte_aligned=True)
        assert x == 8

    def test_find_startbit_not_byte_aligned(self):
        a = Tibs("0b0010000100")
        found = a.find("0b1", start=4)
        assert found == 7
        found = a.find("0b1", start=2)
        assert found == 2
        found = a.find("0b1", start=8, byte_aligned=False)
        assert found is None

    def test_find_endbit_not_byte_aligned(self):
        a = Tibs("0b0010010000")
        found = a.find("0b1", end=2, byte_aligned=False)
        assert found is None
        found = a.find("0b1", end=3)
        assert found == 2
        found = a.find("0b1", start=3, end=5, byte_aligned=False)
        assert found is None
        found = a.find("0b1", start=3, end=6)
        assert found == 5

    def test_find_startbit_byte_aligned(self):
        a = Tibs("0xff001122ff0011ff")
        found = a.find("0x22", start=24, byte_aligned=True)
        assert found == 24
        found = a.find("0b111", start=40, byte_aligned=True)
        assert found == 40 + 16

    def test_find_endbit_byte_aligned(self):
        a = Tibs("0xff001122ff0011ff")
        found = a.find("0x22", start=31, byte_aligned=True)
        assert found is None
        found = a.find("0x22", end=32, byte_aligned=True)
        assert found == 24

    def test_find_all(self):
        a = Tibs("0b11111")
        p = a.find_all("0b1")
        assert list(p) == [0, 1, 2, 3, 4]
        p = a.find_all("0b11")
        assert list(p) == [0, 1, 2, 3]
        p = a.find_all("0b10")
        assert list(p) == []
        a = Tibs("0x4733eeff66554747335832434547")
        p = a.find_all("0x47", byte_aligned=True)
        assert list(p) == [0, 6 * 8, 7 * 8, 13 * 8]
        p = a.find_all("0x4733", byte_aligned=True)
        assert list(p) == [0, 7 * 8]
        a = Tibs("0b1001001001001001001")
        p = a.find_all("0b1001", byte_aligned=False)
        assert list(p) == [0, 3, 6, 9, 12, 15]

    def test_find_all_generator(self):
        a = Tibs("0xff1ff4512345ff1234ff12ff")
        p = a.find_all("0xff", byte_aligned=True)
        assert next(p) == 0
        assert next(p) == 6 * 8
        assert next(p) == 9 * 8
        assert next(p) == 11 * 8
        with pytest.raises(StopIteration):
            _ = next(p)

    def test_contains(self):
        a = Tibs("0b1") + "0x0001dead0001"
        assert "0xdead" in a
        assert "0xfeed" not in a

    def test_repr(self):
        bls = ["", "0b1", "0o5", "0x43412424f41", "0b00101001010101"]
        for bs in bls:
            a = Tibs(bs)
            b = eval(a.__repr__())
            assert a == b
        a = Tibs("0b1")
        assert repr(a).splitlines()[0] == "Tibs('0b1')"
        a += "0b11"
        assert repr(a).splitlines()[0] == "Tibs('0b111')"
        a += "0b1"
        assert repr(a).splitlines()[0] == "Tibs('0xf')"

    def test_iter(self):
        a = Tibs("0b001010")
        b = Tibs()
        for bit in a:
            b = b + Tibs.from_bools([bit])
        assert a == b

    def test_non_zero_bits_at_end(self):
        a = Tibs.from_bytes(b"\xff")[:5]
        b = Tibs("0b00")
        a += b
        assert a == "0b1111100"
        assert (a + [0]).to_bytes() == b"\xf8"
        with pytest.raises(AttributeError):
            _ = a.bytes
        with pytest.raises(ValueError):
            _ = a.to_bytes()

    def test_slice_step(self):
        a = Tibs("0x3")
        b = a[::1]
        assert a == b
        assert a[2:4:1] == "0b11"
        assert a[0:2:1] == "0b00"
        assert a[:3] == "0o1"

        a = Tibs("0x0011223344556677")
        assert a[-8:] == "0x77"
        assert a[:-24] == "0x0011223344"
        assert a[-1000:-24] == "0x0011223344"

    def test_interesting_slice_step(self):
        a = Tibs("0b0011000111")
        assert a[7:3:-1] == "0b1000"
        assert a[9:2:-1] == "0b1110001"
        assert a[8:2:-2] == "0b100"
        assert a[100:-20:-3] == "0b1010"
        assert a[100:-20:-1] == "0b1110001100"
        assert a[10:2:-1] == "0b1110001"
        assert a[100:2:-1] == "0b1110001"

    def test_reverse_with_slice(self):
        a = Tibs("0x0012ff")
        b = a.to_mutibs()
        b.reverse()
        assert a == "0x0012ff"
        assert b == "0xff4800"
        a = a[8:16].to_mutibs()
        a.reverse()
        assert a == "0x48"

    def test_cut(self):
        a = Tibs("0x00112233445")
        b = list(a.chunks(8))
        assert b == ["0x00", "0x11", "0x22", "0x33", "0x44", "0x5"]
        b = list(a[8:16].chunks(4))
        assert b == ["0x1", "0x1"]
        b = list(a[0:44].chunks(4, 4))
        assert b == ["0x0", "0x0", "0x1", "0x1"]
        a = Tibs()
        b = list(a.chunks(10))
        assert not b

    def test_cut_errors(self):
        a = Tibs("0b1")
        with pytest.raises(ValueError):
            _ = a.chunks(0)
        with pytest.raises(ValueError):
            _ = a.chunks(-2)
        with pytest.raises(ValueError):
            _ = a.chunks(1, count=-1)

    def test_cut_problem(self):
        s = Tibs("0x1234")
        for n in list(s.chunks(4)):
            s = n + s
        assert s == "0x43211234"

    def test_join_functions(self):
        a = Tibs.from_joined(["0xa", "0xb", "0b1111"])
        assert a == "0xabf"

    def test_difficult_prepends(self):
        a = Tibs("0b1101011")
        b = Tibs()
        for i in range(10):
            b = a + b
        assert b == a * 10

    def test_reverse_bytes(self):
        a = Mutibs("0x123456")
        a.byte_swap()
        assert a == "0x563412"
        b = a + "0b1"
        with pytest.raises(ValueError):
            b.byte_swap()
        a = Mutibs("0x54")
        a = a.byte_swap()
        assert a == "0x54"
        a = Mutibs()
        a = a.byte_swap()
        assert not a

    def test_startswith(self):
        a = Tibs()
        assert a.starts_with(Tibs())
        assert not a.starts_with("0b0")
        a = Tibs("0x12ff")
        assert a.starts_with("0x1")
        assert a.starts_with("0b0001001")
        assert a.starts_with("0x12ff")
        assert not a.starts_with("0x12ff, 0b1")
        assert not a.starts_with("0x2")

    def test_startswith_start_end(self):
        s = Tibs("0x123456")
        assert s[4:].starts_with("0x234")
        assert not s[:11].starts_with("0x123")
        assert s[:12].starts_with("0x123")
        assert s[8:16].starts_with("0x34")
        assert not s[7:16].starts_with("0x34")
        assert not s[9:16].starts_with("0x34")
        assert not s[8:15].starts_with("0x34")

    def test_endswith(self):
        a = Tibs()
        assert a.ends_with("")
        assert not a.ends_with(Tibs("0b1"))
        a = Tibs("0xf2341")
        assert a.ends_with("0x41")
        assert a.ends_with("0b001")
        assert a.ends_with("0xf2341")
        assert not a.ends_with("0x1f2341")
        assert not a.ends_with("0o34")

    def test_endswith_start_end(self):
        s = Tibs("0x123456")
        assert s[:16].ends_with("0x234")
        assert not s[13:].ends_with("0x456")
        assert s[12:].ends_with("0x456")
        assert s[8:16].ends_with("0x34")
        assert s[7:16].ends_with("0x34")
        assert not s[9:16].ends_with("0x34")
        assert not s[8:15].ends_with("0x34")

    def test_const_bit_stream_hashability(self):
        a = Tibs("0x1")
        b = Tibs("0x2")
        c = Tibs("0x1")
        s = {a, b, c}
        assert len(s) == 2
        assert hash(a) == hash(c)

    def test_hash_edge_cases(self):
        a = Tibs("0xabcd")
        b = Tibs("0xabcd")
        c = b[1:]
        assert hash(a) == hash(b)
        assert hash(a) != hash(c)

    def test_const_bits_copy(self):
        a = Tibs("0xabc")
        b = copy.copy(a)
        assert id(a) == id(b)


class TestSet:
    def test_set(self):
        a = Tibs.from_zeros(16)
        a = a.to_mutibs().set(True, 0)
        assert a == Mutibs("0b10000000 00000000")
        a.set(1, 15)
        assert a == "0b10000000 00000001"
        b = a[4:12]
        b = b.set(True, 1)
        assert b == "0b01000000"
        b = b.set(True, -1)
        assert b == "0b01000001"
        b = b.set(1, -8)
        assert b == "0b11000001"
        with pytest.raises(IndexError):
            _ = b.set(True, -9)
        with pytest.raises(IndexError):
            _ = b.set(True, 8)

    def test_set_negative_index(self):
        a = Mutibs.from_string('0b0110000000')
        a.set(1, -1)
        assert a.to_bin() == "0110000001"
        a.set(1, [-1, -10])
        assert a.to_bin() == "1110000001"
        with pytest.raises(IndexError):
            a.set(1, [-11])

    def test_set_list(self):
        a = Tibs.from_zeros(18)
        b = a.to_mutibs().set(True, range(18))
        assert b.all()
        b.set(False, range(18))
        assert not b.any()

    def test_unset(self):
        a = Mutibs.from_ones(16)
        a.set(False, 0)
        b = ~a
        assert b == "0b10000000 00000000"
        a.set(0, 15)
        assert ~a == "0b10000000 00000001"
        b = a[4:12]
        b.set(False, 1)
        assert ~b == "0b01000000"
        b.set(False, -1)
        assert ~b == "0b01000001"
        b.set(False, -8)
        assert ~b == "0b11000001"
        with pytest.raises(IndexError):
            b.set(False, -9)
        with pytest.raises(IndexError):
            b.set(False, 8)


class TestInvert:
    def test_invert_bits(self):
        a = Mutibs("0b111000")
        a = a.invert(range(len(a)))
        assert a == "0b000111"
        a = a.invert([0, 1, -1])
        assert a == "0b110110"

    def test_invert_whole_bit_stream(self):
        a = Mutibs("0b11011")
        a = a.invert()
        assert a == "0b00100"

    def test_invert_single_bit(self):
        a = Mutibs("0b000001")
        a = a.invert(0)
        assert a.to_bin() == "100001"
        a = a.invert(-1)
        assert a.to_bin() == "100000"

    def test_invert_errors(self):
        a = Mutibs.from_zeros(10)
        with pytest.raises(IndexError):
            _ = a.invert(10)
        with pytest.raises(IndexError):
            _ = a.invert(-11)
        with pytest.raises(IndexError):
            _ = a.invert([1, 2, 10])

    def test_ior(self):
        a = Tibs("0b1101001")
        a |= "0b1110000"
        assert a == "0b1111001"
        b = a[2:]
        c = a[1:-1]
        b |= c
        assert c == "0b11100"
        assert b == "0b11101"

    def test_iand(self):
        a = Tibs("0b0101010101000")
        a &= "0b1111110000000"
        assert a == "0b0101010000000"

    def test_ixor(self):
        a = Tibs("0b11001100110011")
        a ^= "0b11111100000010"
        assert a == "0b00110000110001"

    def test_logical_inplace_errors(self):
        a = Tibs.from_zeros(4)
        with pytest.raises(ValueError):
            a |= "0b111"
        with pytest.raises(ValueError):
            a &= "0b111"
        with pytest.raises(ValueError):
            a ^= "0b111"


class TestAllAndAny:
    def test_all(self):
        a = Tibs("0b111")
        assert a.all()

    def test_any(self):
        a = Tibs("0b10011011")
        assert a.any()

    def test_all_false(self):
        a = Tibs("0b0010011101")
        assert not a.all()

    def test_any_false(self):
        a = Tibs("0b0000")
        assert not a.any()

    def test_any_empty_bitstring(self):
        a = Tibs()
        assert not a.any()  # Python any function returns False for empty iterables.

    def test_all_empty_bit_stream(self):
        a = Tibs()
        assert a.all()  # Python all function returns True for empty iterables.

    def test_any_whole_bitstring(self):
        a = Tibs("0xfff")
        assert a.any()

    def test_all_whole_bitstring(self):
        a = Tibs("0xfff")
        assert a.all()

    ###################


class TestMoreMisc:

    def test_ror(self):
        a = Tibs("0b11001").to_mutibs()
        a.ror(0)
        assert a == "0b11001"
        a.ror(1)
        assert a == "0b11100"
        a.ror(5)
        assert a == "0b11100"
        a.ror(101)
        assert a == "0b01110"
        a = Mutibs("0b1")
        a.ror(1000000)
        assert a == "0b1"

    def test_ror_errors(self):
        a = Mutibs()
        with pytest.raises(ValueError):
            a.ror(0)
        a += "0b001"
        with pytest.raises(ValueError):
            a.ror(-1)

    def test_rol(self):
        a = Mutibs("0b11001")
        a.rol(0)
        assert a == "0b11001"
        a.rol(1)
        assert a == "0b10011"
        a = a.rol(5)
        assert a == "0b10011"
        a.rol(101)
        assert a == "0b00111"
        a = Tibs("0b1")
        a = a.to_mutibs().rol(1000000)
        assert a == "0b1"

    def test_rol_errors(self):
        a = Mutibs()
        with pytest.raises(ValueError):
            a.rol(0)
        a += "0b001"
        with pytest.raises(ValueError):
            a.rol(-1)

    def test_init_with_zeros(self):
        a = Tibs.from_zeros(0)
        assert not a
        a = Tibs.from_zeros(1)
        assert a == "0b0"
        with pytest.raises(ValueError):
            _ = Tibs.from_zeros(-1)
        with pytest.raises(TypeError):
            a += 10

    def test_add_verses_in_place_add(self):
        a1 = Tibs("0xabc")
        b1 = a1
        a1 += "0xdef"
        assert a1 == "0xabcdef"
        assert b1 == "0xabc"

    def test_and_verses_in_place_and(self):
        a1 = Tibs("0xabc")
        b1 = a1
        a1 &= "0xf0f"
        assert a1 == "0xa0c"
        assert b1 == "0xabc"

    def test_or_verses_in_place_or(self):
        a1 = Tibs("0xabc")
        b1 = a1
        a1 |= "0xf0f"
        assert a1 == "0xfbf"
        assert b1 == "0xabc"

    def test_xor_verses_in_place_xor(self):
        a1 = Tibs("0xabc")
        b1 = a1
        a1 ^= "0xf0f"
        assert a1 == "0x5b3"
        assert b1 == "0xabc"

    def test_mul_verses_in_place_mul(self):
        a1 = Tibs("0xabc")
        b1 = a1
        a1 *= 3
        assert a1 == "0xabcabcabc"
        assert b1 == "0xabc"

    def test_lshift_verses_in_place_lshift(self):
        a1 = Tibs("0xabc")
        b1 = a1
        a1 <<= 4
        assert a1 == "0xbc0"
        assert b1 == "0xabc"

    def test_rshift_verses_in_place_rshift(self):
        a1 = Tibs("0xabc")
        b1 = a1
        a1 >>= 4
        assert a1 == "0x0ab"
        assert b1 == "0xabc"


class TestBugs:
    def test_bug_in_replace(self):
        s = Mutibs("0x00112233")
        s = s.replace("0x22", "0xffff", start=8, byte_aligned=True)
        assert s == "0x0011ffff33"
        s = Mutibs("0x0123412341234")
        s = s.replace("0x23", "0xf", start=9, byte_aligned=True)
        assert s == "0x012341f41f4"

    def test_function_negative_indices(self):
        # insert
        s = Mutibs("0b0111")
        s = s.insert(-1, "0b0")
        assert s == "0b01101"

        # overwrite
        t = Mutibs("0x77ab9988c7bf")
        t[-20: -20 + 12] = "0x666"
        assert t == "0x77ab998666bf"

        # find
        found = t.find("0x998", start=-32, byte_aligned=True)
        assert found == 16
        found = t.find("0x988", end=-21, byte_aligned=True)
        assert found is None
        found = t.find("0x998", end=-20, byte_aligned=True)
        assert found == 16

        # find_all
        s = Tibs("0x1234151f")
        li = list(s.find_all("0x1", start=-15))
        assert li == [24]
        li = list(s.find_all("0x1", start=-16, byte_aligned=True))
        assert li == [16, 24]
        li = list(s.find_all("0x1", end=-5, byte_aligned=True))
        assert li == [0, 16]
        li = list(s.find_all("0x1", end=-4, byte_aligned=True))
        assert li == [0, 16, 24]

        # rfind
        found = (s.rfind("0x1f", end=-1))
        assert found is None
        found = s.rfind("0x12", start=-31)
        assert found is None

        # chunks
        s = Tibs("0x12345")
        li = list(s[-12:-4].chunks(4))
        assert li == ["0x3", "0x4"]

        # startswith
        s = Mutibs("0xfe0012fe1200fe")
        assert s[-16:].starts_with("0x00f")
        assert s[:-40].starts_with("0xfe00")
        assert not s[:-41].starts_with("0xfe00")

        # endswith
        assert s[-16:].ends_with("0x00fe")
        assert not s[-15:].ends_with("0x00fe")
        assert not s[-1:].ends_with("0x00fe")
        assert s[:-4].ends_with("0x00f")

        # replace
        s.replace("0xfe", "", end=-1)
        assert s == "0x00121200fe"
        s.replace("0x00", "", start=-24)
        assert s == "0x001212fe"

    def test_rotate_start_and_end(self):
        a = Mutibs("0b110100001")
        a.rol(1, 3, 6)
        assert a == "0b110001001"
        a.ror(1, start=-4)
        assert a == "0b110001100"
        a.rol(202, end=-5)
        assert a == "0b001101100"
        a.ror(3, end=4)
        assert a == "0b011001100"
        with pytest.raises(ValueError):
            a.rol(5, start=-4, end=-6)

    def test_byte_swap_int(self):
        s = Mutibs("0xf234567f")
        s.byte_swap(1)
        assert s == "0xf234567f"
        s.byte_swap(2)
        assert s == "0x34f27f56"
        s.byte_swap(2)
        assert s == "0xf234567f"
        with pytest.raises(ValueError):
            s.byte_swap(3)

    def test_byte_swap_errors(self):
        s = Mutibs("0x0011223344556677")
        with pytest.raises(TypeError):
            s.byte_swap("z")
        with pytest.raises(ValueError):
            s.byte_swap(-1)
        with pytest.raises(TypeError):
            s.byte_swap([-1])
        with pytest.raises(TypeError):
            s.byte_swap(5.4)


def test_count():
    a = Tibs("0xf0f")
    assert a.count(True) == 8
    assert a.count(False) == 4

    b = Tibs()
    assert b.count(True) == 0
    assert b.count(False) == 0

    a = Tibs("0xff0120ff")
    b = a[1:-1]
    assert b.count(1) == 16
    assert b.count(0) == 14


def test_overwrite_with_self():
    s = Mutibs("0b1101")
    s[:] = s
    assert s == "0b1101"


def test_byte_swap():
    b = Tibs.from_bytes(b"\x01\x02\x03\x04")
    c = b.to_mutibs().byte_swap()
    assert c == "0x04030201"


def test_overlapping_bits():
    a = Tibs('0x00fff0')
    zeros = a[0:8]
    x = a[4:16]
    y = x[1:9]
    assert a == "0x00fff0"
    assert zeros == "0x00"
    assert x == "0x0ff"
    assert y == Tibs("0b00011111")
    _ = ~y
    _ = y.to_mutibs().set(0, [0, 1, 2, 3, 4, 5, 6, 7])
    _ = y.to_mutibs().byte_swap()
    _ = y.to_mutibs().ror(1)
    _ = y.to_mutibs().rol(1)
    assert a == "0x00fff0"
    assert zeros == "0x00"
    assert x == "0x0ff"
    assert y == Tibs("0b00011111")
    y = ~y
    assert y == Tibs("0b11100000")
    y = y.to_mutibs().set(0, [2, 3]).to_tibs()
    y = y.to_mutibs().byte_swap().to_tibs()
    y = y.to_mutibs().ror(2)
    y = y.rol(1)
    assert a == "0x00fff0"
    assert zeros == "0x00"
    assert x == "0x0ff"
    assert y == Tibs("0b01100000")


def test_mutable_freeze():
    a = Mutibs('0x0000')
    b = a.to_tibs()
    assert isinstance(b, Tibs)
    assert a == b
    a.set(1, -1)
    assert a == '0x0001'
    assert b.to_hex() == '0000'


def test_del_unavailability():
    a = Tibs('0xff')
    with pytest.raises(TypeError):
        del a[:]
